import requests

from . import API_ENDPOINT_EVENT

import simplekml
import numpy as np


class Bolide():
    """Represent a bright fireball reported at https://neo-bolide.ndc.nasa.gov

    Parameters
    ----------
    eventid : str
        Unique identifier of the event as used by https://neo-bolide.ndc.nasa.gov.
    """
    def __init__(self, eventid):
        self.eventid = eventid
        self.json = self._load_json(eventid)['data'][0]
        self.nSatellites = len(self.json['attachments'])

    def _load_json(self, eventid):
        """Returns a dictionary containing the data for the bolide."""
        url = f"{API_ENDPOINT_EVENT}/{eventid}"
        r = requests.get(url)
        return r.json()

    @property
    def detectedBy(self):
        return self.json['detectedBy']

    @property
    def howFound(self):
        return self.json['howFound']

    @property
    def confidenceRating(self):
        return self.json['confidenceRating']

    @property
    def _id(self):
        return self.json['_id']

    @property
    def attachments(self):
        return self.json['attachments']

    @property
    def netCDFFilename(self):
        # There appears to be a hex number in front of each filename
        # Strip this off
        filenames = []
        for idx in range(self.nSatellites):
            # Make sure we aredealing with a list of strings, it used to be a string
            if type(self.json['attachments'][idx]['netCdfFilename']) != list:
                raise Exception('We are expecting netCdfFilename to be a list of length 1. We got something else')
            for filename in self.json['attachments'][idx]['netCdfFilename']:
                realFilenameIdx = filename.find('_') + 1
                filenames.append(filename[realFilenameIdx:])
        return filenames

    @property
    def satellite(self):
        return [self.json['attachments'][idx]['platformId']
                for idx in range(self.nSatellites)]

    @property
    def platformId(self):
        return self.satellite

    @property
    def geodata(self):
        return [self.json['attachments'][idx]['geoData']
                for idx in range(self.nSatellites)]

    @property
    def longitudes(self):
        return [
                    [x['location']['coordinates'][0]
                    for x in self.geodata[idx]]
                for idx in range(self.nSatellites)]

    @property
    def latitudes(self):
        return [
                    [x['location']['coordinates'][1]
                    for x in self.geodata[idx]]
                for idx in range(self.nSatellites)]

    @property
    def times(self):
        return [[x['time'] for x in self.geodata[idx]]
                for idx in range(self.nSatellites)]

    @property
    def energies(self):
        return [[x['energy'] for x in self.geodata[idx]]
                for idx in range(self.nSatellites)]
    
    def to_lightcurve(self, idx=0):
        """Returns the energies as a LightCurve object."""
        import lightkurve as lk
        return lk.LightCurve(time=self.times[idx],
                             flux=self.energies[idx],
                             targetid=self.eventid[idx])

    def save_kml(self, file_name = ''):

        data_count = len(self.latitudes[0])
        alts = np.linspace(80e3,30e3,data_count)
        lats = self.latitudes[0]
        lons = self.longitudes[0]
        energies = self.energies[0]
        log_energy_mean = np.log10(np.mean(energies))
        log_energies = np.log10(energies)

        self.sizing_scales = (log_energies / log_energy_mean)**5.0

        if not file_name:
            file_name = './kml_file.kml'
        
        kml = simplekml.Kml()
        
        for n, point in enumerate(zip(lons, lats, alts)):
            pnt = kml.newpoint(coords=[point])
            pnt.style.iconstyle.icon.href = 'http://maps.google.com/mapfiles/kml/shapes/placemark_circle.png'
            pnt.style.iconstyle.color = simplekml.Color.red
            pnt.style.iconstyle.scale = self.sizing_scales[n]
            pnt.altitudemode = simplekml.AltitudeMode.relativetoground
        
        kml.save(file_name)



from cartopy.crs import Geostationary, EqualEarth
from .constants import GOES_E_LON, GOES_W_LON, GLM_STEREO_MIDPOINT, FY4A_LON

class DefaultCRS(EqualEarth):
    """The default CoordinateReferenceSystem.

    An Equal-Earth projection from the point between GOES-W and GOES-E
    """
    def __init__(self):
        super().__init__(central_longitude=GLM_STEREO_MIDPOINT)

class GOES_E(Geostationary):
    """Cartopy CoordinateReferenceSystem representing the GOES-East perspective"""
    def __init__(self):
        super().__init__(central_longitude=GOES_E_LON)


class GOES_W(Geostationary):
    """Cartopy CoordinateReferenceSystem representing the GOES-West perspective"""
    def __init__(self):
        super().__init__(central_longitude=GOES_W_LON)


class FY4A(Geostationary):
    """Cartopy CoordinateReferenceSystem representing the Fengyun-4A perspective"""
    def __init__(self):
        super().__init__(central_longitude=FY4A_LON)



from shapely.geometry import Point


class Wrapper():
    """Wrapper class for loading from pickled pipeline data"""
    def __init__(self):
        pass


def make_points(lons, lats):
    """Make Shapely Point objects from lists of longitudes and latitudes"""
    coords = zip(lons, lats)
    points = [Point(coord[0], coord[1]) for coord in coords]
    return points


def reconcile_input(user_input, defaults):
    for key, value in defaults.items():
        if key not in user_input:
            user_input[key] = value

    return user_input


def str_to_list(string):
    string = string.strip('[]')
    if string == '':
        return []
    return string.split(', ')


def youtube_photometry(video):
    """Returns a LightCurve object containing the total grayscale intensity
    of a youtube video over time."""
    # These imports are not used anywhere else, so let's have them be optional/local:
    import cv2
    import numpy as np
    import lightkurve as lk
    from pytube import YouTube

    url = f'https://youtu.be/{video}'
    path = YouTube(url).streams.first().download()
    cap = cv2.VideoCapture(path)
    flux = []
    status = True
    while status:
        status, frame = cap.read()
        if status:
            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            flux.append(np.nansum(gray))
    return lk.LightCurve(flux=flux)



import os
from datetime import datetime, timedelta
from pytz import timezone
from warnings import warn, filterwarnings
from tqdm import tqdm

import numpy as np
import pandas as pd
import pickle

from geopandas import GeoDataFrame

import matplotlib.pyplot as plt
import matplotlib.dates as mdates

from . import MPLSTYLE, ROOT_PATH
from .utils import reconcile_input
from .sources import glm_website, usg, pipeline, gmn, csv, remote

_FIRST_COLS = ['datetime', 'longitude', 'latitude', 'source', 'detectedBy',
               'confidenceRating', 'confidence', 'lightcurveStructure', 'energy',
               'integrated_energy_g16', 'integrated_energy_g17', 'integrated_energy_g18',
               'peak_energy_g16', 'peak_energy_g17', 'peak_energy_g18',
               'peak_energy_cat_g16', 'peak_energy_cat_g17', 'peak_energy_cat_g18',
               'impact-e', 'alt', 'vel']

utc = timezone('UTC')

class BolideDataFrame(GeoDataFrame):
    """
    Subclass of GeoPandas `~geopandas.GeoDataFrame` with additional bolide-specific methods.

    Parameters
    ----------
    source : str
        Specifies the source for the initialized. Can be:

        - ``'glm'``: initialize from neo-bolide-ndc.nasa.gov data
        - ``'glm-pipeline'`` to initialize from ZODB database files from the GLM detection pipeline.
        - ``'pickle'``: initialize from a pickled GeoDataFrame
        - ``'csv'``: initialize from a .csv file
        - ``'usg'``: initialize from US Government data at cneos.jpl.nasa.gov/fireballs/
        - ``'gmn'``: initialize from Global Meteor Networ data at globalmeteornetwork.org/data/
    files : str, list

        Specifies files to be used depending on source.

        - For ``'pickle'``, specifies the filename of the pickled object.
        - For ``'csv'``, specifies the filename of the csv.
        - For ``'glm-pipeline'``, specifies the filename(s) of the database file(s)
    url : str
        Specifies remote url to be used with ``'remote'``.
    date : str, date, datetime
        Specifies a date when using Global Meteor Network data.
        str can be either yyyy-mm or yyyy-mm-dd.

    annotate : bool
        Whether or not to add additional metadata.
    rearrange : bool
        Whether or not to rearrange the columns, if coming from a CSV or Pickle.
    """
    def __init__(self, *args, **kwargs):

        if len(args)==0 and len(kwargs)==0:
            kwargs['source'] = 'glm'
        if 'source' not in kwargs:
            return super().__init__(*args, **kwargs)

        defaults = {'files': None, 'annotate': True, 'rearrange': False}
        kwargs = reconcile_input(kwargs, defaults)
        source = kwargs['source']
        files = kwargs['files']
        annotate = kwargs['annotate']
        rearrange = kwargs['rearrange']

        # input standardization
        source = source.lower()
        if type(files) is str:
            files = [files]

        # input validation
        valid_sources = ['website', 'glm', 'usg', 'pickle', 'gmn', 'csv', 'pipeline',
                         'glm-pipeline', 'usg-orbits', 'glm-orbits', 'remote']
        if source not in valid_sources:
            raise ValueError("Source \""+str(source)+"\" is unsupported. Please use one of "+str(valid_sources))

        if source in ['pickle', 'csv', 'glm-pipeline', 'pipeline'] and files is None:
            raise ValueError("Files must be specified for the given source \""+source+"\"")

        if source in ['pickle', 'csv', 'glm-pipeline', 'pipeline'] and not all([os.path.isfile(f) for f in files]):
            paths = "\n".join([os.path.abspath(f) for f in files])
            raise ValueError("At least one of the files specified does not exist.\
                             Here are their absolute paths:\n"+paths)

        if source in ['pickle', 'csv'] and len(files) > 1:
            warn("More than one file given for source \""+source+"\". Only the first one will be used.")

        # Initialize differently based on source.
        # Each if statement creates a GeoDataFrame with the EPSG:4326 CRS

        if source in ['website', 'glm']:
            source = 'glm'
            init_gdf = glm_website()

        elif source == 'usg':
            init_gdf = usg()

        elif source == 'gmn':
            if 'date' not in kwargs:
                raise ValueError('Must specify a yyyy-mm or yyyy-mm-dd date to use GMN data')
            init_gdf = gmn(kwargs['date'], loc_mode='begin')

        elif source == 'pickle':
            with open(files[0], 'rb') as pkl:
                init_gdf = pickle.load(pkl)

        elif source == 'csv':
            init_gdf = csv(files[0])

        elif source == 'remote':
            if 'url' not in kwargs:
                raise ValueError('Must specify a url with url=...')
            url = kwargs['url']
            init_gdf = remote(url)

        elif source == 'usg-orbits':
            init_gdf = remote('https://aozerov.com/data/usg-orbits.csv')
            annotate = False

        # elif source == 'glm-orbits':
        #     init_gdf = csv(file=ROOT_PATH+'/../notebooks/glm-orbits.csv')
        #     annotate = False

        elif source in ['glm-pipeline', 'pipeline']:
            source = 'glm-pipeline'
            if 'min_confidence' in kwargs:
                min_confidence = kwargs['min_confidence']
            else:
                min_confidence = 0
            init_gdf = pipeline(files=files, min_confidence=min_confidence)

        init_gdf['source'] = source

        # rearrange columns, respecting original order if csv or pickle
        if source not in ['csv', 'pickle'] or rearrange is True:
            first_cols = [col for col in _FIRST_COLS if col in init_gdf.columns]
            other_cols = [col for col in init_gdf.columns if col not in first_cols]
            init_gdf = init_gdf[first_cols + other_cols]

        # initialize the super-class (GeoDataFrame) using the created init_gdf
        super().__init__(init_gdf)

        # add additional metadata to the bolides
        if annotate:
            self.annotate()

        descriptions = pd.read_csv(ROOT_PATH+'/metadata/columns.csv', index_col='column')
        self.descriptions = descriptions[descriptions.sources.isin([source, 'all'])]

    def annotate(self):
        """Add metadata to bolide detections"""
        from .astro_utils import get_phase, get_solarhour, get_sun_alt

        # lunar phase
        self['phase'] = [get_phase(dt) for dt in self['datetime']]
        # moon fullness
        self['moon_fullness'] = -np.abs(self['phase']-0.5)*2+1
        # solar hour
        self['solarhour'] = [get_solarhour(data[0], data[1]) for data in zip(self['datetime'], self['longitude'])]
        # solar altitude
        sun_alt = np.array([get_sun_alt(dt=row['datetime'].to_pydatetime(), lat=row['latitude'], lon=row['longitude']) for _, row in self.iterrows()])
        self['sun_alt_obs'] = sun_alt[:, 0]
        self['sun_alt_app'] = sun_alt[:, 1]

        self['date_retrieved'] = datetime.now()

    def describe(self, key=None):
        """Describe a variable of the BolideDataFrame

        Parameters
        ----------
        key : str, iterable, or None
            The variables to be described.
            If str, describes that variable.
            If iteralbe, describe all variable strings in the iterable.
            If None, describe all variables in the BolideDataFrame.

        Returns
        -------
        str :
            A string describing the variable(s).
        """

        if type(key) is str:
            key = [key]
        to_describe = self.columns if key is None else key
        if 'source' in self.columns and len(self)>0:
            source = self['source'].iloc[0]
        else:
            source = ''
        sources = ['all', source]
        for column in to_describe:
            description = self.descriptions['description'][column] if column in self.descriptions.index else ""
            print(column+":", description)

    def filter_date(self, start=None, end=None, inplace=False):
        """Filter bolides by date.

        Filters the `~BolideDataFrame` using dates given in ISO format.
        `start` or `end` can be left unspecified to only bound the dates on one end.

        Parameters
        ----------
        start, end: str
            ISO-format strings that can be read by `~datetime.datetime.fromisoformat`.
            If the timezone is not specified, it is assumed to be in UTC.
        inplace: bool
            If True, the `~BolideDataFrame` of this method is altered. If False, it is not, so the returned
            `~BolideDataFrame` must be used.

        Returns
        -------
        `~BolideDataFrame`
        """
        new_bdf = self

        # drop data before start date, if specified
        if start is not None:
            dt = datetime.fromisoformat(start)

            # if timezone not given, make the datetime timezone-aware
            # with UTC assumed as the timezone
            if dt.tzname() == None:
                dt = utc.localize(dt)

            to_drop = self.datetime < dt
            new_bdf = new_bdf.drop(new_bdf.index[to_drop], inplace=inplace)
        if inplace:
            new_bdf = self
        # drop data after end date, if specified
        if end is not None:
            dt = datetime.fromisoformat(end)

            # if timezone not given, make the datetime timezone-aware
            # with UTC assumed as the timezone
            if dt.tzname() == None:
                dt = utc.localize(dt)

            to_drop = new_bdf.datetime > dt
            new_bdf = new_bdf.drop(new_bdf.index[to_drop], inplace=inplace)
        if inplace:
            new_bdf = self

        # force the class, which gets converted to GeoDataFrame at some point
        # in the drop methods above
        force_bdf_class(new_bdf)

        return new_bdf

    def get_closest_by_time(self, datestr, n=1):
        """Get the n bolides closest to a given iso-format date string

        Parameters
        ----------
        datestr : str
            ISO-format string that can be read by `~datetime.datetime.fromisoformat`.
            If the timezone is not specified, it is assumed to be in UTC.
        n : int
            Number of closest detections to return.

        Returns
        -------
        bdf : `~BolideDataFrame`
            The filtered data.

        """

        dt = datetime.fromisoformat(datestr)

        # if timezone not given, make the datetime timezone-aware
        # with UTC assumed as the timezone
        if dt.tzname() == None:
            dt = utc.localize(dt)

        return self.iloc[(self['datetime'] - dt).abs().argsort()].head(n)

    def get_closest_by_loc(self, lon, lat, n=1):
        """Get the n bolides closest to a given longitude and latitude

        Parameters
        ----------
        lon, lat : int
            Longitude and latitude to search for bolides near.
        n : int
            Number of closest detections to return.

        Returns
        -------
        bdf : `~BolideDataFrame`
            The filtered data.

        """
        lon_diff = (self['longitude'] - lon).abs()
        lat_diff = (self['latitude'] - lat).abs()
        tot_diff = lon_diff + lat_diff
        return self.iloc[tot_diff.argsort()].head(n)

    def filter_boundary(self, boundary, intersection=False, interior=True):
        """Filter data to only points within specified boundaries.

        Parameters
        ----------
        boundary: str or list of str
            The boundary (boundaries) to filter the `~BolideDataFrame` by.
            Refer to `~bolides.fov_utils.get_boundary`.
        intersection: bool
            - True: keep detections within the union of all boundaries given.
            - False: keep detections within the intersection of all boundaries given.
        interior: bool
            - True: default behavior.
            - False: keep detections outside the union or intersection of the boundaries, instead of inside.

        Returns
        -------
        `~BolideDataFrame`
            The filtered `~BolideDataFrame`
        """

        import pyproj
        # define Azimuthal Equidistant projection
        aeqd = pyproj.Proj(proj='aeqd', ellps='WGS84', datum='WGS84', lat_0=90, lon_0=0).srs

        # put bdf into the correct CRS
        crs = self.geometry.crs
        bdf = self.to_crs(aeqd)

        from .fov_utils import get_boundary
        final_polygon = get_boundary(boundary, intersection=intersection, collection=False)

        # clip with the polygon, which is in Azimuthal Equidistant
        points_in_poly = [pt.within(final_polygon) for pt in bdf['geometry']]
        if interior:
            bdf = bdf[points_in_poly]
        else:
            bdf = bdf[~points_in_poly]
        # project bdf back to original CRS
        bdf = bdf.to_crs(crs)

        force_bdf_class(bdf)

        return bdf

    def filter_observation(self, sensors, intersection=False):
        """Filter data to only points observable (in time and space) by a given sensor.

        Parameters
        ----------
        sensors: str or list of str
            The sensor(s) to filter the `~BolideDataFrame` by:

            - ``'GLM-16'``: The Geostationary Lightning Mapper aboard GOES-16.
            - ``'GLM-17'``: The Geostationary Lightning Mapper aboard GOES-17.
              Note that the biannual yaw flips of GOES-17 are taken into account.
              The shorthand ``'G16'`` and ``'G17'`` can also be used.

        intersection: bool
            If True, filter for bolides observable by all sensors given.
            If False, filter for bolides observable by any sensors given.

        Returns
        -------
        `~BolideDataFrame`
            The filtered `~BolideDataFrame`
        """

        # input standardization and validation
        if type(sensors) is str:
            sensors = [sensors]
        if type(sensors) is not list or not all([type(s) is str for s in sensors]):
            raise ValueError("Sensors must be a string or list of strings")
        sensors = [sensor.lower() for sensor in sensors]

        valid_sensors = ['g16', 'g17', 'glm-16', 'glm-17']

        indices = []
        for num, sensor in enumerate(sensors):
            filename = sensor
            if sensor in ['g16', 'glm-16']:
                filename = ROOT_PATH + '/data/glm16_obs.csv'
            elif sensor in ['g17', 'glm-17']:
                filename = ROOT_PATH + '/data/glm17_obs.csv'
            else:
                raise ValueError("Invalid sensor \""+sensor+"\". sensors must be in "+str(valid_sensors))

            bdfs = []

            # read csv defining the observation times and FOV
            fov_df = pd.read_csv(filename)

            # for each observation time + FOV, filter_date by the observation
            # times, filter_boundary by the FOV, and append result to the list
            # of BolideDataFrames
            for i in range(len(fov_df)):
                start = None if fov_df.start.isna()[i] else fov_df.start[i]
                end = None if fov_df.end.isna()[i] else fov_df.end[i]
                boundary = fov_df.boundary[i]
                bdfs.append(self.filter_boundary([boundary]).filter_date(start=start, end=end))
            bdf = pd.concat(bdfs)
            if not intersection:
                indices += list(bdf.index)
            elif num == 0:
                indices = list(bdf.index)
            else:
                indices = [idx for idx in indices if idx in list(bdf.index)]

        # get a list of unique indices, and sort them
        indices = list(set(indices))
        indices.sort()

        # get the final filtered BolideDataFrame and return
        filtered = self.loc[indices, :]
        force_bdf_class(filtered)
        return filtered

    def filter_shower(self, shower=None, years=None, padding=1, sdf=None, exclude=False):
        """Filter data to only points observable (in time and space) by a given sensor.

        Parameters
        ----------
        shower: str, int, list of str, or list of int
            The meteor shower(s) to filter by. Can enter either IAU number,
            IAU 3-letter code, or full shower name. Refer to the IAU Meteor
            Data center at https://www.ta3.sk/IAUC22DB/MDC2007/.
        years: int or list of int
            The shower years to include. Default is to filter for all occurrences
            of the shower(s) in the BolideDataFrame
        padding: int
            The number of days to pad around the peak time.
        sdf: ShowerDataFrame
            Optionally provide a ShowerDataFrame to use for filtering
        exclude: bool
            Whether or not to exclude bolides around the given showers. Default is to include.

        Returns
        -------
        `~BolideDataFrame`
            The filtered `~BolideDataFrame`
        """
        if years is None:
            years = list(range(min(self.datetime).year-1, max(self.datetime).year+1))
        elif type(years) is int:
            years = [years]

        if sdf is None:
            if hasattr(self, '_showers'):
                sdf = self._showers
            else:
                from . import ShowerDataFrame
                sdf = ShowerDataFrame()
                self._showers = sdf

        dates = sdf.get_dates(shower, years).datetime
        date_padding = timedelta(days=padding)
        date_ranges = [[d-date_padding, d+date_padding] for d in dates]

        counts = np.sum(np.array([list(self.datetime.between(d[0], d[1])) for d in date_ranges]), axis=0)
        if not exclude:
            good_locs = counts != np.zeros(len(counts))
        else:
            good_locs = counts == np.zeros(len(counts))
        return self[good_locs]

    def plot_detections(self, category=None, *args, **kwargs):
        """Plot detections of bolides.

        Reprojects the geometry of bdf to the crs given, and scatters the points
        on a cartopy map. kwargs are passed through to matplotlib's scatter.

        Parameters
        ----------
        crs : `~cartopy.crs.CRS`
            The map projection to use. Refer to
            https://scitools.org.uk/cartopy/docs/latest/reference/projections.html.
        boundary : str or list of str
            The boundaries to plot.
            Refer to `~bolides.fov_utils.get_boundary`.
        category : str
            The name of a categorical column in the `~BolideDataFrame`
        **kwargs :
            Keyword arguments passed through to `~matplotlib.pyplot.scatter`.

        Other Parameters
        ----------------
        coastlines: bool
            Whether or not to draw coastlines.
        style : str
            The matplotlib style to use. Refer to
            https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html
        boundary_style : dict
            The kwargs to use when plotting the boundary.
            Refer to `~cartopy.mpl.geoaxes.GeoAxes.add_geometries`.
        figsize : tuple
            The size (width, height) of the plotted figure.

        Returns
        -------
        fig : `~matplotlib.pyplot.figure`
        ax : `~cartopy.mpl.geoaxes.GeoAxesSubplot`
        """

        # filter to only detections containing both latitude and longitude
        bdf = self[(~self.latitude.isnull()) & (~self.longitude.isnull())]

        if category is not None:
            category = self[category]

        from .plotting import plot_scatter
        fig, ax = plot_scatter(bdf.latitude, bdf.longitude, category=category, *args, **kwargs)

        return fig, ax

    def plot_interactive(self, *args, **kwargs):
        """Plot an interactive map of bolide detections.

        Parameters
        ----------
        mode : str
            Either ``'earth'`` or ``'radiant'``.
            ``'earth'`` plots locations on the Earth,
            ``'radiant'`` plots locations in the sky.
        projection : str
            The map projection to use. Here is the complete list:
            [``'eckert4'``, ``'goes-e'``, ``'goes-w'``, ``'fy4a'``,
            ``'airy'``, ``'aitoff'``, ``'albers'``, ``'albers usa'``, ``'august'``,
            ``'azimuthal equal area'``, ``'azimuthal equidistant'``, ``'baker'``,
            ``'bertin1953'``, ``'boggs'``, ``'bonne'``, ``'bottomley'``, ``'bromley'``,
            ``'collignon'``, ``'conic conformal'``, ``'conic equal area'``, ``'conic equidistant'``,
            ``'craig'``, ``'craster'``, ``'cylindrical equal area'``,
            ``'cylindrical stereographic'``, ``'eckert1'``, ``'eckert2'``,
            ``'eckert3'``, ``'eckert4'``, ``'eckert5'``, ``'eckert6'``, ``'eisenlohr'``,
            ``'equirectangular'``, ``'fahey'``, ``'foucaut'``, ``'foucaut sinusoidal'``,
            ``'ginzburg4'``, ``'ginzburg5'``, ``'ginzburg6'``,
            ``'ginzburg8'``, ``'ginzburg9'``, ``'gnomonic'``, ``'gringorten'``,
            ``'gringorten quincuncial'``, ``'guyou'``, ``'hammer'``, ``'hill'``,
            ``'homolosine'``, ``'hufnagel'``, ``'hyperelliptical'``,
            ``'kavrayskiy7'``, ``'lagrange'``, ``'larrivee'``, ``'laskowski'``,
            ``'loximuthal'``, ``'mercator'``, ``'miller'``, ``'mollweide'``, ``'mt flat polar parabolic'``,
            ``'mt flat polar quartic'``, ``'mt flat polar sinusoidal'``,
            ``'natural earth'``, ``'natural earth1'``, ``'natural earth2'``,
            ``'nell hammer'``, ``'nicolosi'``, ``'orthographic'``,
            ``'patterson'``, ``'peirce quincuncial'``, ``'polyconic'``,
            ``'rectangular polyconic'``, ``'robinson'``, ``'satellite'``, ``'sinu mollweide'``,
            ``'sinusoidal'``, ``'stereographic'``, ``'times'``,
            ``'transverse mercator'``, ``'van der grinten'``, ``'van der grinten2'``,
            ``'van der grinten3'``, ``'van der grinten4'``,
            ``'wagner4'``, ``'wagner6'``, ``'wiechel'``, ``'winkel tripel'``,
            ``'winkel3'``]
        boundary : str or list of str
            The boundaries to plot.
            Refer to `~bolides.fov_utils.get_boundary`.
        color : str
            The name of a column in the `~BolideDataFrame`
        logscale : bool
            Whether or not to use a logarithmic scale for the colors.
        **kwargs :
            Keyword arguments passed through to `~plotly.express.scatter_geo`.

        Other Parameters
        ----------------
        coastlines: bool
            Whether or not to draw coastlines.
        style : str
            The matplotlib style to use. Refer to
            https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html
        boundary_style : dict
            The kwargs to use when plotting the boundary.
            Refer to `~cartopy.mpl.geoaxes.GeoAxes.add_geometries`.
        figsize : tuple
            The size (width, height) of the plotted figure.
        culture : str
            When mode is ``'radiant'``, the asterism source culture to use. Here is the complete list:
            [``'anutan'``, ``'aztec'``, ``'belarusian'``, ``'boorong'``,
             ``'chinese'``, ``'chinese_contemporary'``, ``'chinese_medieval'``,
             ``'egyptian'``, ``'hawaiian_starlines'``, ``'indian'``,
             ``'inuit'``, ``'japanese_moon_stations'``, ``'korean'``,
             ``'lokono'``, ``'macedonian'``, ``'maori'``, ``'maya'``,
             ``'mongolian'``, ``'navajo'``, ``'norse'``, ``'northern_andes'``,
             ``'romanian'``, ``'russian_siberian'``, ``'sami'``,
             ``'sardinian'``, ``'tongan'``, ``'tukano'``, ``'tupi'``,
             ``'western'``, ``'western_SnT'``, ``'western_hlad'``,
             ``'western_rey'``]


        Returns
        -------
        fig : `~matplotlib.pyplot.figure`
        ax : `~cartopy.mpl.geoaxes.GeoAxesSubplot`
        """
        from .plotting import plot_interactive
        return plot_interactive(self, *args, **kwargs)

    def plot_density(self, *args, **kwargs):
        """Plot bolide detection density.

        Density is computed using scikit-learn's `~sklearn.neighbors.KernelDensity`
        using the haversine distance metric (as the data is in longitude and latitude)
        and gaussian kernel by default.
        It is then gridded, projected, and plotted.

        Parameters
        ----------
        crs : `~cartopy.crs.CRS`
            The map projection to use. Refer to
            https://scitools.org.uk/cartopy/docs/latest/reference/projections.html.
        bandwidth : float
            The bandwidth of the Kernel Density Estimator, in degrees.
        boundary : str or list of str
            The boundaries to plot and clip the density by.
            Refer to `~bolides.fov_utils.get_boundary`.
        n_levels : int
            Number of discrete density levels to plot.
        lat_resolution, lon_resolution : ints
            The number of discrete latitude and longitude levels when gridding the density.
        **kwargs :
            Keyword arguments passed through to `~cartopy.mpl.geoaxes.GeoAxes.contourf`.

        Other Parameters
        ----------------
        coastlines : bool
            Whether or not to draw coastlines.
        style : str
            The matplotlib style to use. Refer to
            https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html
        boundary_style : dict
            The kwargs to use when plotting the boundary.
            Refer to `~cartopy.mpl.geoaxes.GeoAxes.add_geometries`.
        kde_params : dict
            The kwargs to pass to `~sklearn.neighbors.KernelDensity`.
            Note that 'metric' is not allowed to be specified, as haversine
            is the only valid metric.
        figsize : tuple
            The size (width, height) of the plotted figure.

        Returns
        -------
        fig : `~matplotlib.pyplot.figure`
        ax : `~cartopy.mpl.geoaxes.GeoAxesSubplot`
        """

        # filter to only detections containing both latitude and longitude
        bdf = self[(~self.latitude.isnull()) & (~self.longitude.isnull())]

        from .plotting import plot_density
        fig, ax = plot_density(bdf.latitude, bdf.longitude, *args, **kwargs)

        return fig, ax

    def plot_dates(self, freq='1D', logscale=False,
                   start=None, end=None,
                   figsize=(10, 3), style=MPLSTYLE,
                   showers=None, line_style={}, **kwargs):
        """Plot the number of bolides over time.

        Parameters
        ----------
        freq : str
            The binning frequency. Can be strings like ``'2D'``, ``'1M'``, etc.
            Refer to https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
        start, end: strs
            Optional arguments to filter by date.
            Must be ISO-format strings that can be read by `~datetime.datetime.fromisoformat`.
        logscale : bool
            Whether or not to use a log-scale for the y-axis
        **kwargs :
            Keyword arguments passed through to `~matplotlib.axes.Axes.bar`.

        Other Parameters
        ----------------
        style : str
            The matplotlib style to use. Refer to
            https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html
        showers : ShowerDataFrame
            A ShowerDataFrame to use for the shower data. By default, showers are pulled
            from the established showers list at the IAU meter data center.
        figsize : tuple
            The size (width, height) of the plotted figure.
        line_style : dict
            The matplotlib style arguments for the vertical shower lines.

        Returns
        -------
        fig : `~matplotlib.pyplot.figure`
        ax : `~matplotlib.axes.Axes`
        """
        from pandas.plotting import register_matplotlib_converters
        register_matplotlib_converters()

        # filter date to given start and end times
        bdf = self.filter_date(start, end, inplace=False)
        datetimes = bdf.datetime

        # get counts according to the given frequency
        index = bdf.groupby(pd.Grouper(key='datetime', freq=freq)).count().index

        if showers is not None:

            # a ShowerDataFrame is stored in the _showers attribute, so that
            # users can make multiple plots without constantly recreating it
            if hasattr(self, '_showers'):
                sdf = self._showers
            else:
                from . import ShowerDataFrame
                sdf = ShowerDataFrame()
                self._showers = sdf
            date_info = sdf.get_dates(showers, range(min(bdf.datetime).year-1, max(bdf.datetime).year+1))
            shower_names = date_info['shower name']
            shower_dates = date_info.datetime

        # with the given matplotlib style,
        with plt.style.context(style):

            # initialize figure and set major, minor ticks and formats
            fig, ax = plt.subplots(figsize=figsize, dpi=300)
            ax.xaxis.set_major_locator(mdates.YearLocator())
            ax.xaxis.set_minor_locator(mdates.MonthLocator())
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%Y'))
            ax.xaxis.set_minor_formatter(mdates.DateFormatter('%m'))

            # make minor ticks (months) small, pad major ticks (years)
            plt.tick_params(axis='x', which='minor', labelsize=5, pad=0, labeltop=True)
            plt.tick_params(axis='x', which='major', pad=3, labeltop=True)

            # set x-label, adding year data if data is only within one year
            # (and hence no year data on x-axis)
            ax.set_xlabel("Date")
            if min(bdf.datetime).year == max(bdf.datetime).year:
                ax.set_xlabel("Date (month in "+str(min(bdf.datetime).year)+")")
            ax.set_ylabel("# Events")

            plt.hist(datetimes, bins=index, **kwargs)

            # set x-limits to data limits
            x_min = datetime.fromisoformat(start) if start is not None else min(bdf.datetime)
            x_max = datetime.fromisoformat(end) if end is not None else max(bdf.datetime)
            plt.xlim(x_min, x_max)

            if showers is not None:
                names = list(shower_names)
                unique = list(shower_names.unique())
                import matplotlib.cm as cm
                cmap = cm.Set2

                defaults = {'linestyle': 'dashed', 'linewidth': 1,
                            'alpha': 0.5}
                line_kwargs = reconcile_input(line_style, defaults)

                for name, date in zip(shower_names, shower_dates):
                    plt.axvline(x=date, color=cmap(unique.index(name)/8),
                                label=name, **line_kwargs)
                # remove extra labels for repeat showers
                for i, p in enumerate(ax.get_lines()):
                    if p.get_label() in names[:i]:
                        p.set_label('_' + p.get_label())
                plt.legend(loc='upper right')

            # make y-axis logscale if specified
            if logscale:
                ax.set_yscale('log')

        return fig, ax

    def add_website_data(self, ids=None):
        """Pull light curve and integrated energy data from neo-bolide.ndc.nasa.gov.

        Downloads light curve data from neo-bolide.ndc.nasa.gov, placing it
        into the BolideDataFrame as a column of `~lightkurve.LightCurveCollection` objects.
        The column is named ``'lightcurves'``. Integrated energy data is also downloaded
        for different satellites and placed into columns starting with ``'energy_'``.
        This method will only work with BolideDataFrames having an ``_id`` column containing
        IDs from GLM bolide detections at neo-bolide.ndc.nasa.gov.

        Parameters
        ----------
        ids : list of str
            Optional list of strings representing the bolide ID's that additional data is
            needed for. If not used, data is added for every bolide in the BolideDataFrame.
        """

        # if there is no _id column, we can't associate the website data to
        # the bolides in the BolideDataFrame
        assert '_id' in self.columns, "BolideDataFrame must have an '_id' column"

        # check that ids is not a single string
        assert type(ids) is not str, "Input must be must be a list"

        # make ids a List if it is an iterable
        if hasattr(ids, '__iter__'):
            ids = list(ids)

            # check that all the given ids actually exist
            assert all([_id in list(self._id) for _id in ids]), "All given IDs must exist in the BolideDataFrame"

        # if no ids are given, we assume that we will obtain website data for each
        # event in the BolideDataFrame
        if ids == None:
            ids = list(self._id)

        # at this point, unless the input was not iterable, ids should be a list
        assert type(ids) is list, "Input must be a list"

        # get the data from the website by passing the ids
        from .sources import glm_website_event
        data = glm_website_event(ids)

        # initialize empty columns in the BolideDataFrame
        cols = data.keys()
        for col in cols:
            if col == 'lightcurves':
                self[col] = None
            else:
                self[col] = np.nan

        # iterate through the list of ids, plugging the obtained data
        # into the appropriate rows in the BolideDataFrame
        for i, _id in enumerate(ids):

            # find the row containing the id
            idx = self.index[np.argmax(self._id == _id)]

            # enter the data
            for col in cols:
                self.loc[idx, col] = data[col][i]

    # TODO: match on a column other than _id?
    def augment(self, new_data, time_limit=300, score_limit=5, intersection=False, outer=False):
        """Augment BolideDataFrame with data from another source

        Parameters
        ----------
        new_data : `~BolideDataFrame`
            The source of the new data.
        time_limit : int
            Maximum time difference (seconds) for two bolides to possibly be the same
        score_limit : int
            Maximum score (time_limit * difference in latlon) for two bolides to
            possibly be the same.
        intersection : bool
            - True: only keep detections that appear in both sets of data.
            - False: keep all detections that appear in the first set.
        outer : bool
            - True: keep detections from both sets of data.
            - False: keep all detections that appear in the first set.

        Returns
        -------
        bdf : `~BolideDataFrame`
        """

        # make _id column if doesn't exist
        if "_id" not in self.columns:
            self["_id"] = np.arange(len(self))

        # iterate through rows of self, computing distance and identifying
        # detections with detections in the other data if certain closeness
        # criteria based on time and distance are met
        ids = [""] * len(new_data)
        for num, row in tqdm(self.iterrows(), "Augmenting data", total=len(self)):
            deltas = row['datetime']-new_data['datetime']
            s_deltas = np.abs([delta.total_seconds() for delta in deltas])
            closest = np.argmin(s_deltas)
            if s_deltas[closest] < time_limit:
                lat_diff = abs(row['latitude']-new_data['latitude'].iloc[closest])
                lon_diff = abs(row['longitude'] % 360 - new_data['longitude'].iloc[closest] % 360)
                geo_diff = lat_diff+lon_diff
                s_diff = s_deltas[closest]
                score = geo_diff * s_diff
                if score < score_limit:
                    ids[closest] = row['_id']
        new_data['_id'] = ids
        # merge the data. If taking the intersection, only keep rows in
        # original BolideDataFrame that have a corresponding detection in
        # the other source
        if intersection:
            merged = self.merge(new_data, 'inner', on='_id')
        elif outer:
            merged = self.merge(new_data, 'outer', on='_id')
        else:
            merged = self.merge(new_data, 'left', on='_id')

        # rename the required columns
        for col in merged.columns:
            if col.endswith('_x') and col[:-2]+"_y" in merged.columns:
                merged[col[:-2]] = merged[col]
                del merged[col]

        # force class
        merged.__class__ = BolideDataFrame

        return merged

    def __setattr__(self, attr, val):
        # since BolideDataFrame can have non-column attributes, we can ignore
        # Pandas' warnings about setting attributes
        filterwarnings("ignore", message="Pandas doesn't allow columns to be created via a new attribute name")
        return super().__setattr__(attr, val)

    # override GeoPandas' __getitem__ to force the class to remain BolideDataFrame
    def __getitem__(self, key):
        result = super().__getitem__(key)
        force_bdf_class(result)
        return result

    @property
    def _constructor(self):
        return BolideDataFrame

    # define how the BolideDataFrame renders in an IPython notebook
    def _repr_html_(self):

        # want to show all columns
        with pd.option_context('display.max_columns', None):
            # get the default representation from Pandas
            df_rep = super()._repr_html_()

        good_sources = ['glm', 'usg', 'glm-pipeline']

        attribution = ""

        # get the attribution HTML file
        if 'source' in self.columns and len(self) > 0:
            source = self['source'].iloc[0]
            if source in good_sources:
                with open(ROOT_PATH + '/metadata/' + source + '.html', 'r') as f:
                    attribution = f.read()
            if 'source_y' in self.columns and self['source_y'].iloc[0] in good_sources:
                source = self['source_y'].iloc[0]
                attribution += "Augmented with "
                with open(ROOT_PATH + '/metadata/' + source + '.html', 'r') as f:
                    attribution += f.read()
        return attribution + df_rep


def force_bdf_class(bdf):
    if isinstance(bdf, GeoDataFrame):
        bdf.__class__ = BolideDataFrame



import os
import requests

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import pandas as pd

from . import Bolide
from . import API_ENDPOINT_EVENTLIST


class BolideList():

    def __init__(self):
        self.json = self._load_json()

    def _load_json(self):
        """Returns a dictionary containing all events."""
        r = requests.get(API_ENDPOINT_EVENTLIST)
        return r.json()

    @property
    def ids(self):
        return [event["_id"] for event in self.json['data']]

    def __len__(self):
        return len(self.ids)

    def __getitem__(self, idx):
        return Bolide(self.ids[idx])

    def to_pandas(self):
        """Returns a pandas DataFrame summarizing all bolides."""
        df = pd.DataFrame(self.json['data'])
        df["datetime"] = df["datetime"].astype("datetime64")
        return df

    def plot_dates(self, year=2019):
        """Plots the number of bolides over time."""
        from pandas.plotting import register_matplotlib_converters
        register_matplotlib_converters()
        df = self.to_pandas()
        mask = (df.datetime > f'{year}-01-01') & (df.datetime < f'{year+1}-01-01')
        counts = df[mask].groupby(pd.Grouper(key='datetime', freq='1D')).count()
        plt.style.use("ggplot")
        fig, ax = plt.subplots(figsize=(10, 3), dpi=300)
        ax.xaxis.set_major_locator(mdates.MonthLocator())
        ax.xaxis.set_major_formatter(mdates.DateFormatter('%b %d'))
        ax.set_xlabel(year)
        ax.set_ylabel("# Events")
        ax.bar(counts.index, counts._id);
        return ax

    def plot_map(self):
        """Plots the spatial distribution of bolides using basemap."""
        from mpl_toolkits.basemap import Basemap
        fig = plt.figure(figsize=(9*1.618, 9))
        ax = fig.add_axes([0.02, 0.03, 0.96, 0.89])
        ax.set_facecolor('white')
        m = Basemap(projection='kav7',
                    lon_0=-90, lat_0=0,
                    resolution="l", fix_aspect=False)
        m.drawcountries(color='#7f8c8d', linewidth=0.8)
        m.drawstates(color='#bdc3c7', linewidth=0.5)
        m.drawcoastlines(color='#7f8c8d', linewidth=0.8)
        m.fillcontinents('#ecf0f1', zorder=0)
        df = self.to_pandas()
        x, y = m(df.longitude.values, df.latitude.values)
        m.scatter(x, y, marker="o", color="red", edgecolor='black',
                lw=0.4, s=15, zorder=999)
        plt.show()




class AMSBolideList():

    def __init__(self, year=2019):
        self.json = self._load_json(year)

    def _load_json(self, year):
        key = os.environ.get('AMS_API_KEY')
        url = f"https://www.amsmeteors.org/members/api/open_api/get_events?api_key={key}&year={year}&format=json"
        r = requests.get(url)
        return r.json()

    @property
    def events(self):
        return [self.json['result'][key] for key in self.json['result']]

    def to_pandas(self):
        """Returns a pandas DataFrame summarizing all bolides."""
        import pandas as pd
        df = pd.DataFrame(self.json['result']).transpose()
        df["avg_date_utc"] = df["avg_date_utc"].astype("datetime64")
        return df



import pandas as pd
import requests
import numpy as np
from . import ROOT_PATH


class ShowerDataFrame(pd.DataFrame):
    """
    Subclass of Pandas `~pandas.DataFrame` with additional meteor-shower-specific methods.

    Parameters
    ----------
    source : str
        Specifies the source for the initialized. Can be:

        - ``'established', 'all', 'working'``: initialize from different sets
          of data offered at the IAU Meteor Data Center, https://www.ta3.sk/IAUC22DB/MDC2007/
        - ``'csv'``: initialize from a .csv file

    file : str
        Specifies files to be used if the source is ``'csv'``
    """
    def __init__(self, *args, **kwargs):
        if len(args)==0 and len(kwargs)==0:
            kwargs['source'] = 'established'
        if 'source' not in kwargs:
            return super().__init__(*args, **kwargs)
        if 'file' not in kwargs:
            kwargs['file'] = None
        source = kwargs['source']
        file = kwargs['file']
        if source == 'csv':
            df = pd.read_csv(file, index_col=0, keep_default_na=False, na_values='')
        else:
            url = 'https://www.ta3.sk/IAUC22DB/MDC2007/Etc/stream'+source+'data.txt'
            r = requests.get(url)
            start_line = 0
            for num, line in enumerate(r.text.splitlines()):
                if line[0] not in [':', '+']:
                    start_line = num
                    break
            column_line = r.text.splitlines()[start_line-3]
            import re
            columns = re.split(r" {2,}", column_line)[1:-1]
            data_lines = r.text.splitlines()[start_line:]
            data = '\n'.join(data_lines)
            import io
            csv_io = io.StringIO(data)
            df = pd.read_csv(csv_io, sep="|", header=None)
            df.columns = columns
            for col in df.columns:
                if df[col].dtype == 'O':
                    df[col] = df[col].str.strip()

        column_translation = {'inc': 'i', 'Ra': 'ra', 'De': 'dec'}
        df = df.rename(columns=column_translation)
        df['source'] = 'iau'

        numeric_cols = ['a', 'e', 'peri', 'node', 'i']
        for col in numeric_cols:
            df[col] = pd.to_numeric(df[col], errors='coerce')
        super().__init__(df)

    def plot_orbits(self, date='2000-01-01T12:00:00',
                    use_3d=False, interactive=False, num_points=150):
        """Plot an interactive map of shower orbits.

        Parameters
        ----------
        date : str
            The date at which to plot the data, in ISO format.
        use_3d : bool
            Whether or not to make a 3D plot using Plotly
        interactive : bool
            Whether or not to make an interactive plot using Plotly

        Returns
        -------
        plotter : `~poliastro.plotting.core.OrbitPlotter3D` or `~poliastro.plotting.core.OrbitPlotter2D`
        """

        if use_3d:
            interactive = True

        from poliastro.bodies import Sun
        from astropy import units as u
        from poliastro.twobody import Orbit

        from datetime import datetime
        dt = datetime.fromisoformat(date)
        from astropy.time import Time
        epoch = Time(dt, scale='tdb').tdb

        from poliastro.plotting.misc import _plot_bodies
        from poliastro.plotting.interactive import OrbitPlotter2D, OrbitPlotter3D
        from poliastro.plotting.static import StaticOrbitPlotter

        from poliastro.frames import Planes

        if use_3d:
            plotter = OrbitPlotter3D(plane=Planes.EARTH_ECLIPTIC)
        elif interactive:
            plotter = OrbitPlotter2D(plane=Planes.EARTH_ECLIPTIC)
        else:
            plotter = StaticOrbitPlotter(plane=Planes.EARTH_ECLIPTIC)

        from warnings import simplefilter
        import erfa
        simplefilter("ignore", erfa.core.ErfaWarning)

        plotter._num_points = num_points
        plotter.set_attractor(Sun)

        _plot_bodies(plotter, epoch=epoch)

        has_data = all([col in self.columns for col in ['a', 'e', 'i', 'node', 'peri']])

        if has_data:
            sdf = self.dropna(subset=['a', 'e', 'i', 'node', 'peri'])

            import warnings
            if use_3d and len(sdf) > 50:
                warnings.warn('This will plot a lot of orbits and may crash your browser.\n'
                              'If you are afraid of this happening, please do now use show()')


            for num, row in sdf.iterrows():
                a = row['a'] * u.AU
                ecc = row['e'] * u.one
                inc = row['i'] * u.deg
                raan = row['node'] * u.deg
                argp = row['peri'] * u.deg
                nu = 5 * u.deg
                try:
                    orb = Orbit.from_classical(Sun, a, ecc, inc, raan, argp, nu, plane=Planes.EARTH_ECLIPTIC)
                except ValueError:
                    continue
                plotter.plot(orb, label=row['shower name'])

        if use_3d:
            fig = plotter._figure
            fig.update_layout(margin={"r": 0, "t": 0, "l": 0, "b": 0}, height=900)
            axis_dict = {'showgrid': False, 'zeroline': False, 'visible': False}
            fig.update_layout(scene={'xaxis': axis_dict, 'yaxis': axis_dict, 'zaxis': axis_dict})

            # replace spheres with points and shorten labels
            import plotly.graph_objects as go
            good_data = []
            planets = ['Sun', 'Mercury', 'Venus', 'Earth', 'Mars', 'Jupiter', 'Saturn', 'Uranus', 'Neptune']
            i = 0
            sun_added = False
            while(i < len(fig.data)):
                data = fig.data[i]
                orig_name = data.name
                sun = data.name == 'Sun'
                if not sun:
                    data.name = " (".join(data.name.split(' (')[1:])[:-1]

                planet = any([data.name.__contains__(p) for p in planets])
                if planet and type(data) is go.Surface:
                    name = orig_name
                    x = np.mean(data.x)
                    y = np.mean(data.y)
                    z = np.mean(data.z)
                    color = data.colorscale[0][1]
                    fig.add_trace(go.Scatter3d(x=[x], y=[y], z=[z], name=name))
                    fig.update_traces(marker=dict(color=color, size=4),
                                      selector=dict(name=name, type='scatter3d'),
                                      showlegend=False)
                if not planet:
                    fig.update_traces(line_dash='dot', selector=dict(name=data.name, type='scatter3d'))
                if type(data) is not go.Surface and (not sun or not sun_added):
                    good_data.append(data)
                    if sun:
                        sun_added = True
                i += 1
            fig.data = good_data

            min_x = min([min(d.x) for d in fig.data])
            min_y = min([min(d.y) for d in fig.data])
            min_z = min([min(d.z) for d in fig.data])
            max_x = max([max(d.x) for d in fig.data])
            max_y = max([max(d.y) for d in fig.data])
            max_z = max([max(d.z) for d in fig.data])

            a_x = max_x-min_x
            a_y = max_y-min_y
            a_z = max_z-min_z
            c_x = min_x + a_x/2
            c_y = min_y + a_y/2
            c_z = min_z + a_z/2

            longest = max([a_x, a_y, a_z])
            fig.update_scenes(aspectratio=dict(x=a_x/longest, y=a_y/longest, z=a_z/longest),
                              aspectmode="manual")

            x = (0-c_x)/a_x * a_x/longest
            y = (0-c_y)/a_y * a_y/longest
            z = (0-c_z)/a_z * a_z/longest
            fig.update_layout(scene_camera=dict(up=dict(x=0, y=0, z=1),
                              center=dict(x=x, y=y, z=z)))

        return plotter

    def get_dates(self, showers, years):
        if type(years) is int:
            years = [years]
        if type(showers) in [str, int]:
            showers = [showers]
        showers = [str(s) for s in showers]
        if showers[0].isdigit():
            col = 'IAUNo'
            showers = [int(s) for s in showers]
        elif len(showers[0]) == 3:
            col = 'Code'
        else:
            col = 'shower name'
        sdf = self[self[col].isin(showers)]
        import warnings
        if len(sdf) == 0:
            warnings.warn('No showers with '+col+'in'+str(showers)+'.')

        sdf = sdf[sdf.activity == 'annual']
        sdf = sdf.dropna(subset=['LaSun'])
        num = len(sdf)
        sdf = pd.concat([sdf]*len(years), ignore_index=True)
        years = np.repeat(years, num)

        lons = sdf.LaSun
        from .astro_utils import sol_lon_to_datetime
        dts = []
        for lon, year in zip(lons, years):
            dts.append(sol_lon_to_datetime(lon, year))
        return pd.DataFrame({'Code': sdf.Code,
                             'shower name': sdf['shower name'],
                             'IAUNo': sdf.IAUNo,
                             'datetime': dts})

    def __setattr__(self, attr, val):
        from warnings import filterwarnings
        filterwarnings("ignore", message="Pandas doesn't allow columns to be created via a new attribute name")
        return super().__setattr__(attr, val)

    def __getitem__(self, key):
        result = super().__getitem__(key)
        force_showers_class(result)
        return result

    @property
    def _constructor(self):
        return ShowerDataFrame

    def _repr_html_(self):
        with pd.option_context('display.max_columns', None):
            return super()._repr_html_()


def force_showers_class(showers):
    if isinstance(showers, pd.DataFrame):
        showers.__class__ = ShowerDataFrame



from math import degrees, pi, floor, radians, sin
from datetime import datetime
from subprocess import Popen, PIPE
import numpy as np
import ephem
import astropy.units as u
from astropy.coordinates import ICRS, SkyCoord
from astropy.time import Time

from pytz import timezone
utc = timezone('UTC')


def get_phase(datetime):
    """Get lunar phase (0.01=new moon just happened, 0.99=new moon about to happen)"""
    date = ephem.Date(datetime)
    nnm = ephem.next_new_moon(date)
    pnm = ephem.previous_new_moon(date)

    lunation = (date-pnm)/(nnm-pnm)
    return lunation


def get_solarhour(datetime, lon):
    """Get the hour in solar time given a datetime and longitude"""
    o = ephem.Observer()
    o.date = datetime
    o.long = lon/180 * pi
    sun = ephem.Sun()
    sun.compute(o)
    hour_angle = o.sidereal_time() - sun.ra
    solarhour = ephem.hours(hour_angle+ephem.hours('12:00')).norm/(2*pi) * 24
    return solarhour


def get_sun_alt(dt, lat, lon):
    """Get the solar altitude given a date and location"""
    obs = ephem.Observer()
    obs.lon = str(lon)
    obs.lat = str(lat)
    obs.date = dt
    sun = ephem.Sun()
    sun.compute(obs)
    observed = degrees(sun.alt)
    obs.pressure = 0
    sun.compute(obs)
    apparent = degrees(sun.alt)
    return np.array([observed, apparent])


def vel_to_radiant(dt, vx, vy, vz):
    """Input velocity in ITRS frame, output (uncorrected) radiant in ICRS frame"""

    time = Time(dt)
    # input negatives of coordinates because we want the direction they're coming from
    c = SkyCoord(x=-vx, y=-vy, z=-vz, representation_type='cartesian', frame='itrs', obstime=time)
    radec = c.transform_to(ICRS)
    return radec.ra.value, radec.dec.value


def geocentric_to_ecliptic(ra, dec):
    """Given ra and dec, compute ecliptic latitude and longitude"""

    c = SkyCoord(ra=ra*u.degree, dec=dec*u.degree, frame='icrs')
    lat = c.barycentrictrueecliptic.lat.value
    lon = c.barycentrictrueecliptic.lon.value
    return lat, lon


def calc_orbit(dt, v, vx, vy, vz, lat, lon, alt, wmpl_path='python'):
    """
    Input velocity vector and position in ITRS frame, output orbital elements.

    wmpl_path must be a callable Python instance which has the
    WesternMeteorPyLib installed.
    """

    keys = ['ra', 'dec', 'LaSun', 'a', 'e', 'i', 'peri', 'node', 'Pi', 'b', 'q', 'f', 'M', 'Q', 'n', 'T']

    time = Time(dt)
    # input negatives of coordinates because we want the direction they're coming from
    c = SkyCoord(x=-vx, y=-vy, z=-vz, representation_type='cartesian', frame='itrs', obstime=time)
    radec = c.transform_to(ICRS)
    ra = radec.ra.value
    dec = radec.dec.value
    datestr = dt.strftime('%Y%m%d-%H%M%S.0')
    args = f'-r {ra} -d {dec} -v {v} -t {datestr} -a {lat} -o {lon} -e {alt} -s'.split()
    process = Popen([wmpl_path, "-m", "wmpl.Trajectory.Orbit"]+args, stdout=PIPE, stderr=PIPE)
    stdout, stderr = process.communicate()
    stdout = stdout.decode()
    contains_data = stdout.__contains__('Orbit:')
    if not contains_data:
        data = [np.nan]*len(keys)
    else:
        lines = stdout.split('\n')
        new_radiant_line = np.argmax([line.__contains__('Radiant (geocentric, J2000)') for line in lines])
        ra = float(lines[new_radiant_line+1].split()[2].strip('+'))
        dec = float(lines[new_radiant_line+2].split()[2].strip('+'))
        data = [ra, dec]

        orbit_line = np.argmax([line.__contains__('Orbit:') for line in lines])
        for i in range(14):
            if i == 0:
                idx = 3
            else:
                idx = 2
            data.append(float(lines[orbit_line+1+i].split()[idx]))
    data_dict = dict(zip(keys, data))
    return data_dict


def sol_lon_to_datetime(lon, year):
    """Given solar longitude and year, compute datetime"""

    JD = sol_lon_to_jd(lon, year)
    t = Time(JD, format='jd', scale='utc')
    dt = t.datetime

    # make the UTC datetime timezone-aware
    utc = timezone('UTC')
    dt = utc.localize(dt)

    return dt


def sol_lon_to_jd(lon, year):
    """Get the Julian Day given a solar longitude.

    An algorithm for computing the Julian Day given a solar longitude,
    as described in:
    Low-Precision Formulae for Calculating Julian Day from Solar Longitude,
    E. Ofek, WGN 2000.
    https://ui.adsabs.harvard.edu/abs/2000JIMO...28..176O.

    This algorithm is approximate (see paper for details)
    but is good enough for plotting and filtering
    """

    lon = radians(lon)
    Y = year
    M = floor(lon/360 * 12)+3
    D = 1
    dt = datetime(Y, M, D, 0, 0, 0)
    t = Time(dt, format='datetime', scale='utc')
    JD_approx = t.jd

    M0 = 2451182.24736
    M1 = 365.25963575
    A1 = 1.94330
    phi1 = -1.798135
    A2 = 0.013053
    phi2 = 2.634232
    B1 = 78.1927
    B2 = 58.13165
    P2 = -0.0000089408


    N = year - 2000
    JD0 = M0 + M1*N
    TDelta = A1*sin(lon+phi1) + A2*sin(2*lon+phi2) + B1 + B2*lon + P2*(JD_approx - 2451545)
    if abs(JD_approx-JD0-TDelta) > 50:
        TDelta += 365.2596
    JD = JD0 + TDelta
    return JD



# This module contains high-level plotting tools for use with the BolieDataFrame class, 
# but also accessible for general use by other packages.

import pandas as pd
import numpy as np
from warnings import warn, filterwarnings

import matplotlib.pyplot as plt

from geopandas import GeoDataFrame
from shapely.geometry import Point

from . import MPLSTYLE, ROOT_PATH
from .utils import reconcile_input

def plot_scatter(
        latitude, 
        longitude, 
        crs=None,
        category=None, 
        boundary=None, 
        coastlines=True, 
        style=MPLSTYLE,
        boundary_style={}, 
        figsize=(10, 5),
        fig=None,
        ax=None,
        **kwargs):
    """Plot detections of bolides.

    Reprojects the geometry of bdf to the crs given, and scatters the points
    on a cartopy map. kwargs are passed through to matplotlib's scatter.

    Parameters
    ----------
    latitude : np.array
        The array of latitudes (degrees) to plot the density.
        Using EPSG:4326 - WGS 84, latitude/longitude coordinate system
        sized N
    longitude : np.array
        The array of longitudes (degrees) to plot the density.
        Using EPSG:4326 - WGS 84, latitude/longitude coordinate system
        sized N
    crs : `~cartopy.crs.CRS`
        The map projection to use. Refer to
        https://scitools.org.uk/cartopy/docs/latest/reference/projections.html.
    category : str list
        The category label to color each datum in the scatter plot
        sized N
    boundary : str or list of str
        The boundaries to plot.
        Refer to `~bolides.fov_utils.get_boundary`.
    **kwargs :
        Keyword arguments passed through to `~matplotlib.pyplot.scatter`.

    Other Parameters
    ----------------
    coastlines: bool
        Whether or not to draw coastlines.
    style : str
        The matplotlib style to use. Refer to
        https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html
    boundary_style : dict
        The kwargs to use when plotting the boundary.
        Refer to `~cartopy.mpl.geoaxes.GeoAxes.add_geometries`.
    figsize : tuple
        The size (width, height) of the plotted figure.
    c : np.array or list
        A ana array to use for the mappint to color each point
    color : str
        A single matplotlib color string
    fig : `~matplotlib.pyplot.figure`
        If passed then use this figure, otherwise, create one
        Use genearate_plot to return the fig and ax
        Note: Both fig and ax must be passed or neither
    ax : `~cartopy.mpl.geoaxes.GeoAxesSubplot`
        If passed then use this axis, otherwise, create one
        Use genearate_plot to return the fig and ax
        Note: Both fig and ax must be passed or neither

    Returns
    -------
    fig : `~matplotlib.pyplot.figure`
    ax : `~cartopy.mpl.geoaxes.GeoAxesSubplot`
    """
    
    # Both fig and ax must be passed or neither
    if fig is None != ax is None:
        # XOR gate
        'Both fig and ax must be passed or neither'
    elif fig is None and ax is None:
        fig_and_ax_passed = False
    else:
        fig_and_ax_passed = True

    # Project the latitudes and longitudes to the correct projection
    points = [Point(lon, lat) for lon, lat in zip(longitude, latitude)]
    gdf = GeoDataFrame(geometry=points, crs='epsg:4326')
    
    from .crs import DefaultCRS
    if crs is None:
        crs = DefaultCRS()

    # The cartopy library used by plot_detections currently has many
    # warnings about the shapely library deprecating things...
    # This code suppresses those warnings
    filterwarnings("ignore", message="__len__ for multi-part")
    filterwarnings("ignore", message="Iteration over multi-part")

    import matplotlib.cm as cmx

    # get geopandas projection and reproject dataframe points
    crs_proj4 = crs.proj4_init
    gdf_proj = gdf.to_crs(crs_proj4)
    # filter out rows with no geometry

    # default parameters put into kwargs if not specified by user
    assert not ('c' in kwargs and 'color' in kwargs), "Either 'c' or 'color' may be passed but not both"
    defaults = {'marker': '.', 'color': 'red', 'cmap': plt.get_cmap('viridis')}
    kwargs = reconcile_input(kwargs, defaults)
    if 'c' in kwargs:
        del kwargs['color']
    else:
        del kwargs['cmap']  # We are using a solid color, no color mapping

    good_locs = ~gdf_proj.geometry.is_empty
    x = np.empty(len(gdf_proj))
    y = np.empty(len(gdf_proj))
    points = gdf_proj[good_locs]['geometry']
    x[good_locs] = np.array([p.x for p in points])
    y[good_locs] = np.array([p.y for p in points])

    # using the given style,
    with plt.style.context(style):

        # generate Figure and GeoAxes with the given projection
        if not fig_and_ax_passed:
            fig, ax = generate_plot(crs=crs, figsize=figsize, style=style)
            ax.stock_img()  # plot background map

        # scatter points, passing arguments through

        if category is None:  # if there is no categorical variable specified
            cb = plt.scatter(x, y, **kwargs)
            # if color is determined by a quantitative variable, we add a colorbar
            if 'c' in kwargs:
                plt.colorbar(cb, label=kwargs['c'].name)

        else:  # if there is a categorical variable specified, color points using it
            assert len(category) == len(gdf), 'Variable <category> should be an array of strings the size of the input Lat and Lon'
            unique = category.unique()  # get the unique values of the variable
            import matplotlib.colors as colors
            hot = plt.get_cmap('tab10')
            cNorm = colors.Normalize(vmin=0, vmax=len(unique))
            scalarMap = cmx.ScalarMappable(norm=cNorm, cmap=hot)

            del kwargs['color']  # color kwarg being overridden by categorical variable
            s = kwargs['s'] if 's' in kwargs else None

            # for each unique category, scatter the data with the right color
            for num, label in enumerate(unique):
                idx = category == label
                if s is not None and hasattr(s, '__getitem__'):
                    kwargs['s'] = s[idx]
                ax.scatter(x[idx], y[idx], color=scalarMap.to_rgba(num), label=label, **kwargs)
            plt.legend()

        if coastlines:
            ax.coastlines()  # plot coastlines

        # plot sensor FOV
        from .fov_utils import add_boundary
        if boundary:
            add_boundary(ax, boundary, boundary_style)

    return fig, ax

def plot_density(
        latitude, 
        longitude, 
        crs=None,
        bandwidth=5, 
        boundary=None,
        n_levels=100, 
        lat_resolution=200, 
        lon_resolution=100,
        coastlines=True, 
        style=MPLSTYLE,
        boundary_style={},
        kde_params={}, 
        figsize=(10, 5), 
        title=None,
        fig=None,
        ax=None,
        **kwargs):
    """Plot event density on an Earth projection.

    Density is computed using scikit-learn's `~sklearn.neighbors.KernelDensity`
    using the haversine distance metric (as the data is in longitude and latitude)
    and gaussian kernel by default.
    It is then gridded, projected, and plotted.

    Parameters
    ----------
    latitude : np.array
        The array of latitudes (degrees) to plot the density.
        Using EPSG:4326 - WGS 84, latitude/longitude coordinate system
    longitude : np.array
        The array of longitudes (degrees) to plot the density.
        Using EPSG:4326 - WGS 84, latitude/longitude coordinate system
    crs : `~cartopy.crs.CRS`
        The map projection to use. Refer to
        https://scitools.org.uk/cartopy/docs/latest/reference/projections.html.
    bandwidth : float
        The bandwidth of the Kernel Density Estimator, in degrees.
    boundary : str or list of str
        The boundaries to plot and clip the density by.
        Refer to `~bolides.fov_utils.get_boundary`.
    n_levels : int
        Number of discrete density levels to plot.
    lat_resolution, lon_resolution : ints
        The number of discrete latitude and longitude levels when gridding the density.
    **kwargs :
        Keyword arguments passed through to `~cartopy.mpl.geoaxes.GeoAxes.contourf`.

    Other Parameters
    ----------------
    coastlines : bool
        Whether or not to draw coastlines.
    style : str
        The matplotlib style to use. Refer to
        https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html
    boundary_style : dict
        The kwargs to use when plotting the boundary.
        Refer to `~cartopy.mpl.geoaxes.GeoAxes.add_geometries`.
    kde_params : dict
        The kwargs to pass to `~sklearn.neighbors.KernelDensity`.
        Note that 'metric' is not allowed to be specified, as haversine
        is the only valid metric.
    figsize : tuple
        The size (width, height) of the plotted figure.
    title : str
        The title for the figure, if desired
    fig : `~matplotlib.pyplot.figure`
        If passed then use this figure, otherwise, create one
        Use genearate_plot to return the fig and ax
        Note: Both fig and ax must be passed or neither
    ax : `~cartopy.mpl.geoaxes.GeoAxesSubplot`
        If passed then use this axis, otherwise, create one
        Use genearate_plot to return the fig and ax
        Note: Both fig and ax must be passed or neither

    Returns
    -------
    fig : `~matplotlib.pyplot.figure`
    ax : `~cartopy.mpl.geoaxes.GeoAxesSubplot`

    """
    # Both fig and ax must be passed or neither
    if fig is None != ax is None:
        # XOR gate
        'Both fig and ax must be passed or neither'
    elif fig is None and ax is None:
        fig_and_ax_passed = False
    else:
        fig_and_ax_passed = True

    assert len(latitude) == len(longitude), "latitude and longitude should be the same length."
    num_bolides = len(latitude)

    from .crs import DefaultCRS
    if crs is None:
        crs = DefaultCRS()

    from cartopy import crs as ccrs
    # The cartopy library used by plot_density currently has many
    # warnings about the shapely library deprecating things...
    # This code suppresses those warnings
    filterwarnings("ignore", message="__len__ for multi-part")
    filterwarnings("ignore", message="Iteration over multi-part")

    # get numpy array in the format that KernelDensity likes
    data = np.vstack([np.radians(latitude), np.radians(longitude)]).T

    # set kde_params and validate input
    if 'kernel' not in kde_params:
        kde_params['kernel'] = 'gaussian'
    if 'metric' in kde_params:
        del kde_params['metric']
        warn('Please do not specify metric. Any metric other than haversine (default)\
             will lead to invalid results.')

    from math import radians
    from sklearn.neighbors import KernelDensity
    # create and fit a KDE
    kde = KernelDensity(bandwidth=radians(bandwidth), metric="haversine", **kde_params)
    kde.fit(data)

    # create grid of latitudes and longitudes
    from math import pi
    X, Y = np.meshgrid(np.linspace(-pi, pi, lat_resolution),
                       np.linspace(-pi/2, pi/2, lon_resolution))
    xy = np.vstack([Y.ravel(), X.ravel()]).T

    # compute density at gridpoints
    density_per_steradian = np.exp(kde.score_samples(xy))

    # convert density per steradian to bolides per sqkm
    steradian_per_sqdeg = 1/3282.80635
    earth_sqkm = 510 * 10**6
    earth_sqdeg = 41252.96
    sqdeg_per_sqkm = earth_sqdeg / earth_sqkm
    bolides_per_steradian = num_bolides * density_per_steradian
    bolides_per_sqkm = bolides_per_steradian * steradian_per_sqdeg * sqdeg_per_sqkm

    # prepare for plotting
    Z = bolides_per_sqkm
    Z = Z.reshape(X.shape)
    levels = np.linspace(0, Z.max(), n_levels)
    x = np.degrees(X)
    y = np.degrees(Y)
    z = Z

    # get mask given the boundary
    from .fov_utils import get_mask
    mask = get_mask(np.degrees(xy), boundary).reshape(X.shape)

    # default parameters put into kwargs if not specified by user
    default_cmap = plt.get_cmap('viridis').copy()
    default_cmap.set_under('none')
    defaults = {'alpha': 1, 'antialiased': False, 'cmap': default_cmap}
    kwargs = reconcile_input(kwargs, defaults)

    # using the given style,
    with plt.style.context(style):

        # generate Figure and GeoAxes with the given proejction
        if not fig_and_ax_passed:
            fig, ax = generate_plot(crs=crs, figsize=figsize, style=style)
            ax.stock_img()  # plot background map

        # plot contour, passing arguments through
        filled_c = ax.contourf(x, y, z*mask, levels=levels[1:],
                               transform=ccrs.PlateCarree(), **kwargs)

        # make lines invisible
        for c in filled_c.collections:
            c.set_edgecolor('none')
            c.set_linewidth(0.000000000001)

        if coastlines:
            ax.coastlines()  # plot coastlines

        # plot sensor FOV
        from .fov_utils import add_boundary
        if boundary:
            add_boundary(ax, boundary, boundary_style)

        plt.colorbar(filled_c, alpha=kwargs['alpha'],
                     label='bolide density (km$^{-2}$)')

        if title is not None:
            plt.title(title)

    return fig, ax

def plot_interactive(df, mode='earth', projection="eckert4",
                     boundary=None, color=None, logscale=False,
                     culture='western', reference_plane='ecliptic',
                     **kwargs):

    # TODO: write header and documentation throughout this function

    import plotly.express as px
    import plotly.graph_objects as go
    df = df.copy()
    hover_columns = ['datetime']

    if mode == 'earth':
        loc_cols = ['latitude', 'longitude']
    elif mode == 'radiant':
        loc_cols = ['dec', 'ra']

    if len(df) == 0 or any([col not in df.columns for col in loc_cols]):
        values = [[]] * len(loc_cols)
        df = pd.DataFrame(dict(list(zip(loc_cols, values))))

    if mode == 'radiant' and reference_plane == 'ecliptic':
        from astropy import units as u
        from astropy.coordinates import SkyCoord
        lats = []
        lons = []
        for ra, dec in zip(df['ra'], df['dec']):
            c = SkyCoord(ra=ra*u.degree, dec=dec*u.degree, frame='icrs')
            lats.append(c.barycentrictrueecliptic.lat.value)
            lons.append(c.barycentrictrueecliptic.lon.value)
        df['lat_ecliptic'] = lats
        df['lon_ecliptic'] = lons
        df['lon_ecliptic_neg'] = -df['lon_ecliptic']
        loc_cols = ['lat_ecliptic', 'lon_ecliptic_neg']

    elif mode == 'radiant':
        df['ra_neg'] = -df['ra']
        loc_cols = ['dec', 'ra_neg']

    too_long = ['otherInformation', 'reason', 'description', 'otherDetectingSources', 'status',
                'lastModifiedBy', 'enteredBy', 'submittedBy', 'publishedBy', 'platform', 'rejectedBy',
                'rejectedDate', 'date_retrieved', '__v', 'Remarks', 'References']
    if color not in df.columns:
        color = None

    import numbers

    hover_columns = [col for col in hover_columns if col in df.columns]

    if len(df) > 0:
        for col in df.columns:
            if type(df[col].iloc[0]) in [list, dict]:
                continue
            if all([isinstance(x, numbers.Number) for x in df[col]]) and (col not in too_long):
                hover_columns.append(col)
            elif len(df[col].unique()) < 20 and (col not in too_long):
                hover_columns.append(col)
        for col in hover_columns:
            if col not in loc_cols + [color]:
                if all([isinstance(i, numbers.Number) for i in df[col]]):
                    df[col] = ['%g' % num for num in df[col]]
                else:
                    df[col] = df[col].fillna('')

    hover_columns = [col for col in hover_columns if not col.endswith('_sd')]

    if len(hover_columns) > 30:
        hover_columns = ['datetime', 'longitude', 'latitude', 'ra', 'dec', 'lon_ecliptic', 'lat_ecliptic']
    hover_columns = [col for col in hover_columns if col in df.columns]

    defaults = {'hover_data': hover_columns}
    kwargs = reconcile_input(kwargs, defaults)

    projection = projection.lower()
    proj_name = projection
    if projection in ['goes-e', 'goes-w', 'fy4a']:
        proj_name = 'satellite'
    if logscale and color is not None and all([isinstance(x, numbers.Number) for x in df[color]]):
        fig = px.scatter_geo(df, lat=loc_cols[0], lon=loc_cols[1],
                             color=np.log(df[color]),
                             projection=proj_name, **kwargs)
        fig.update_layout(coloraxis_colorbar={'title': 'log('+color+')'})
    else:
        fig = px.scatter_geo(df, lat=loc_cols[0], lon=loc_cols[1],
                             color=color,
                             projection=proj_name, **kwargs)
    from .constants import GLM_STEREO_MIDPOINT, GOES_E_LON, GOES_W_LON, FY4A_LON
    if len(df) > 0 and df['source'].iloc[0] == 'website' and mode == 'earth':
        fig.update_geos(projection_rotation={'lon': GLM_STEREO_MIDPOINT})
    if projection in ['satellite', 'goes-e', 'goes-w', 'fy4a']:
        distance = 42164/6378  # geostationary orbit distance
        rotation = 0
        if projection == 'goes-e':
            rotation = GOES_E_LON
        elif projection == 'goes-w':
            rotation = GOES_W_LON
        elif projection == 'fy4a':
            rotation = FY4A_LON
        fig.update_geos(projection={'type': 'satellite', 'distance': distance},
                        projection_rotation={'lon': rotation})

    fig.update_layout(margin={"r": 0, "t": 0, "l": 0, "b": 0}, height=900)
    fig.update_geos(landcolor="White", lataxis_showgrid=True, lonaxis_showgrid=True)
    if mode == 'radiant':
        fig.update_geos(landcolor='Black', lakecolor='Black',
                        coastlinecolor='Black', oceancolor='Black', showocean=True,)
    fig.update_traces(marker=dict(size=8))
    if mode == 'earth':
        from bolides.fov_utils import get_boundary
        import pyproj
        import shapely
        if boundary is None:
            boundary = []
        if type(boundary) is str:
            boundary = [boundary]
        polygons = get_boundary(boundary)
        if type(polygons) is not list:
            polygons = [polygons]
        aeqd = pyproj.Proj(proj='aeqd', ellps='WGS84', datum='WGS84', lat_0=90, lon_0=0).srs
        gdf = GeoDataFrame(geometry=polygons, crs=aeqd)
        gdf = gdf.to_crs('epsg:4326')
        polygons = gdf.geometry
        for num, polygon in enumerate(polygons):
            if type(polygon) is shapely.geometry.MultiPolygon:
                polygon_group = list(polygon)
            else:
                polygon_group = [polygon]

            from plotly.colors import qualitative
            for i, p in enumerate(polygon_group):
                lons, lats = p.exterior.coords.xy
                lons = np.array(lons)
                lats = np.array(lats)
                if boundary[num] in ['goes', 'goes-w', 'goes-w-i', 'goes-w-ni']:
                    lons = lons - (lons > 50) * 360
                fig.add_trace(go.Scattergeo(mode="lines", lon=lons, lat=lats,
                                            name=boundary[num], opacity=0.7,
                                            line=dict(color=qualitative.Pastel[num]),
                                            legendgroup=str(num),
                                            showlegend=(i == 0)))
    if mode == 'radiant':
        constellations = pd.read_csv(ROOT_PATH+'/data/constellations/'+culture+'.csv')

        # convert strings to lists
        from .utils import str_to_list
        for col in ['ra', 'dec']:
            constellations[col] = [str_to_list(x) for x in constellations[col]]

        for num, row in constellations.iterrows():
            name = row['name']
            data_x = np.array(row['ra']).astype(float)
            data_y = np.array(row['dec']).astype(float)
            if reference_plane == 'ecliptic':
                from astropy import units as u
                from astropy.coordinates import SkyCoord
                lats = []
                lons = []
                for ra, dec in zip(row['ra'], row['dec']):
                    try:
                        c = SkyCoord(ra=float(ra)*u.degree, dec=float(dec)*u.degree, frame='icrs')
                        ecliptic_coords = c.barycentrictrueecliptic
                        lats.append(ecliptic_coords.lat.value)
                        lons.append(ecliptic_coords.lon.value)
                    except ValueError as e:
                        print(e)
                data_x = np.array(lons)
                data_y = np.array(lats)
            data_x = -data_x
            for n in range(int(len(row['ra'])/2)):
                fig.add_trace(go.Scattergeo(mode="lines", name=name,
                                            hoverinfo='name', hoverlabel={'namelength': -1},
                                            lon=data_x[n*2:(n+1)*2],
                                            lat=data_y[n*2:(n+1)*2],
                                            line=dict(color='White'), showlegend=False))

    return fig

#*************************************************************************************************************
# Helper functions

def generate_plot(crs=None, figsize=(10,5), style=MPLSTYLE):
    """ Helper function to generate the plot figure. 

    This can be used to generate a single figure then call the plotting
    functions in this module repeatedly to superimpose data on the figure.

    Parameters
    ----------
    crs : `~cartopy.crs.CRS`
        The map projection to use. Refer to
        https://scitools.org.uk/cartopy/docs/latest/reference/projections.html.
    figsize : tuple
        The size (width, height) of the plotted figure.
    style : str
        The matplotlib style to use. Refer to
        https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html


    Returns
    -------
    fig : `~matplotlib.pyplot.figure`
    ax : `~cartopy.mpl.geoaxes.GeoAxesSubplot`
    """
    from .crs import DefaultCRS
    if crs is None:
        crs = DefaultCRS()

    with plt.style.context(style):
        fig, ax = plt.subplots(subplot_kw={'projection': crs}, figsize=figsize)
        ax.stock_img()  # plot background map

    return fig, ax



def dict_from_zodb(files, min_confidence):
    import ZODB.FileStorage
    import zc.zlibstorage
    from tqdm import tqdm

    list_of_dicts = []
    for filename in files:

        # use ZODB and zc to open the compressed database
        bolide_db = ZODB.FileStorage.FileStorage(filename)
        compressed = zc.zlibstorage.ZlibStorage(bolide_db)
        large_record_size = 1 << 27
        db = ZODB.DB(compressed, large_record_size=large_record_size)
        connection = db.open()

        # get number of detections
        total = connection.root.n_tot_detections

        # create an iterator that goes through all of the detections
        count = 0
        # loop through iterator, with loading bar
        for key, value in tqdm(connection.root.detections.iteritems(), total=total):
            count += 1
            ID = key
            d = value

            # If the BolideDetection object is not broken, we are in an
            # environment which contains all of the classes that the database
            # requires.
            broken = hasattr(d, '__Broken_state__')
            d = dict_from_obj(d, broken)

            # only add to list of dicts if the confidence is above the threshold
            if d['confidence'] > min_confidence:
                # create dict containing basic data
                data0 = {'_id': ID, 'confidence': d['confidence'],
                         'method': d['confidenceSource'], 'comments': d['howFound'],
                         'yaw_flip_flag': d['yaw_flip_flag']}
                # get dict from the attributes and values of the features object
                features = dict_from_obj(d['features'], broken)

                # get dicts from the attributes and values of the stereo features objects
                stereoFeatures = dict_from_obj(d['stereoFeatures'], broken)
                g16 = dict_from_obj(stereoFeatures['G16'], broken)
                g17 = dict_from_obj(stereoFeatures['G17'], broken)

                g16 = add_key_suffix(g16, '_g16')
                g17 = add_key_suffix(g17, '_g17')

                # Add G18 data only if available in order to support legacy databases
                if 'G18' in stereoFeatures:
                    g18 = dict_from_obj(stereoFeatures['G18'], broken)
                    g18 = add_key_suffix(g18, '_g18')
                    # append a dict combining all of these to the list of dicts
                    list_of_dicts.append(dict(data0, **features, **g16, **g17, **g18))
                else:
                    # append a dict combining all of these to the list of dicts
                    list_of_dicts.append(dict(data0, **features, **g16, **g17))


            # minimize the cache a few times while running
            if count % (total//10) == 0:
                db.cacheMinimize()

    # turn the list of dicts into a dict of lists
    # (under the assumption that all dicts have the same keys)
    dict_of_lists = {key: [dic[key] for dic in list_of_dicts] for key in list_of_dicts[0]}

    return dict_of_lists


# helper function to add a suffix to each key of a dict
def add_key_suffix(dic, suffix):
    new_keys = [key+suffix for key in dic.keys()]
    return dict(zip(new_keys, dic.values()))


# If the object is broken, get a dict using the __Broken_state__ attribute
# If not, get a dict using vars()
def dict_from_obj(obj, broken):
    if broken:
        return obj.__Broken_state__
    else:
        return vars(obj)



import pandas as pd
import numpy as np
from geopandas import GeoDataFrame
import requests
from io import StringIO
from tqdm import tqdm

from . import API_ENDPOINT_EVENTLIST, API_ENDPOINT_EVENT
from .utils import make_points
from datetime import datetime

from lightkurve import LightCurve, LightCurveCollection


def glm_website():

    # load data from website
    json = download(API_ENDPOINT_EVENTLIST, json=True)

    # create DataFrame using JSON data
    df = pd.DataFrame(json['data'])
    df["datetime"] = pd.to_datetime(df["datetime"])

    # add bolide brightness data
    peak_energy_cat_g16 = []
    peak_energy_g16 = []
    peak_energy_cat_g17 = []
    peak_energy_g17 = []
    peak_energy_cat_g18 = []
    peak_energy_g18 = []
    val_cols = {'GLM-16': peak_energy_g16,
                'GLM-17': peak_energy_g17,
                'GLM-18': peak_energy_g18}
    cat_cols = {'GLM-16': peak_energy_cat_g16,
                'GLM-17': peak_energy_cat_g17,
                'GLM-18': peak_energy_cat_g18}

    for brightness in df.brightness:
        for sat in ['GLM-16', 'GLM-17', 'GLM-18']:
            if sat in brightness:
                cat_cols[sat].append(brightness[sat]['category'])
                val_cols[sat].append(brightness[sat]['value'])
            else:
                cat_cols[sat].append("")
                val_cols[sat].append(np.nan)

    df['peak_energy_cat_g16'] = peak_energy_cat_g16
    df['peak_energy_g16'] = peak_energy_g16
    df['peak_energy_cat_g17'] = peak_energy_cat_g17
    df['peak_energy_g17'] = peak_energy_g17
    df['peak_energy_cat_g18'] = peak_energy_cat_g18
    df['peak_energy_g18'] = peak_energy_g18

    del df['brightness']

    gdf = add_geometry(df)

    return gdf


def usg():

    # load data from website
    json = download('https://ssd-api.jpl.nasa.gov/fireball.api?vel-comp=true', json=True)
    data = json['data']
    cols = json['fields']

    # create DataFrame
    df = pd.DataFrame(data, columns=cols)
    df['latitude'] = df['lat'].astype(float) * ((df['lat-dir'] == 'N') * 2 - 1)
    df['longitude'] = df['lon'].astype(float) * ((df['lon-dir'] == 'E') * 2 - 1)
    del df['lat'], df['lon'], df['lat-dir'], df['lon-dir']
    df['datetime'] = [datetime.fromisoformat(date) for date in df['date']]
    # localize to UTC, as that is how USG datetimes are reported
    df['datetime'] = df['datetime'].dt.tz_localize('UTC')
    del df['date']
    numeric_cols = ['energy', 'impact-e', 'alt', 'vel', 'vx', 'vy', 'vz']
    for col in numeric_cols:
        df[col] = df[col].astype(float)

    # compute ra and dec from velocity components
    df['ra'] = np.nan
    df['dec'] = np.nan
    from .astro_utils import vel_to_radiant
    # ignore SettingWithCopyWarning
    with pd.option_context('mode.chained_assignment', None):
        for num, row in df.dropna(subset=['vx', 'vy', 'vz']).iterrows():
            ra, dec = vel_to_radiant(row['datetime'], row['vx'], row['vy'], row['vz'])
            df['ra'][num] = ra
            df['dec'][num] = dec

    first_cols = ['datetime', 'longitude', 'latitude', 'source', 'energy',
                  'impact-e', 'alt', 'vel', 'source']
    first_cols = [col for col in first_cols if col in df.columns]
    other_cols = [col for col in df.columns if col not in first_cols]
    df = df[first_cols + other_cols]

    gdf = add_geometry(df)

    return gdf


def pipeline(files, min_confidence=0):

    from .pipeline_utils import dict_from_zodb
    dict_of_lists = dict_from_zodb(files=files, min_confidence=min_confidence)

    df = pd.DataFrame(dict_of_lists)

    column_translation = {'avgLon': 'longitude', 'avgLat': 'latitude', 'bolideTime': 'datetime',
                          'timeDuration': 'duration', 'goesSatellite': 'detectedBy'}
    df = df.rename(columns=column_translation)
    df['datetime'] = df['datetime'].dt.tz_localize('UTC')

    gdf = add_geometry(df)

    return gdf


def gmn(date, loc_mode='begin'):
    input_date = date
    month = False
    if type(date) is str:
        if len(date) == 7:
            month = True
        elif len(date) == 10:
            pass
        else:
            raise ValueError(f'invalid GMN date {date}')
        datestr = date.replace("-", "")
    else:
        datestr = date.strftime('%Y%m%d')

    base_path = 'https://globalmeteornetwork.org/data/traj_summary_data/'
    if month:
        data_path = f'{base_path}monthly/traj_summary_monthly_{datestr}.txt'
    else:
        # load index page
        from bs4 import BeautifulSoup
        archive_url = "https://globalmeteornetwork.org/data/traj_summary_data/daily/"
        r = requests.get(archive_url)
        soup = BeautifulSoup(r.content, 'html5lib')
        links = soup.findAll('a')

        csv_links = [archive_url + link['href'] for link in links if link['href'].startswith(f'traj_summary_{datestr}')]
        data_path = csv_links[0]
        print(data_path)

    data = download(data_path)
    if data.__contains__('404 Not Found'):
        unit_string = 'monthly' if month else 'daily'
        raise ValueError(f'{input_date} not found in GMN {unit_string} data.')
    buf = StringIO(data)
    buf.readline()
    header = buf.readline().split(';')
    header = [''.join(h.split()) for h in header]
    header[1] = 'jd'
    header[2] = 'datetime'
    header[3] = 'iau_no'
    header[4] = 'iau_code'
    for idx, col in enumerate(header):
        if col == '+/-':
            header[idx] = header[idx-1]+'_sd'
    buf.readline()
    buf.readline()
    df = pd.read_csv(buf, sep=';')
    df.columns = header

    # convert the strings to datetimes in UTC. If the strings have timezone information,
    # they will be converted to UTC, otherwise they are assumed to be in UTC.
    df['datetime'] = pd.to_datetime(df['datetime'], utc=True, format='ISO8601')
    if loc_mode == 'begin':
        df['latitude'] = df.LatBeg
        df['longitude'] = df.LonBeg
    else:
        df['latitude'] = df.LatEnd
        df['longitude'] = df.LonEnd

    column_translation = {'RAgeo': 'ra', 'DECgeo': 'dec', 'RAgeo_sd': 'ra_sd',
                          'DECgeo_sd': 'dec_sd'}
    df = df.rename(columns=column_translation)

    # non_numeric = ['source', '#Uniquetrajectory', 'iau_code', 'Begin', 'Endin', 'Participating']
    # dates = ['datetime', 'date-retrieved']
    # numeric_cols = [col for col in df.columns if col not in non_numeric+dates]
    # for col in numeric_cols:
    #    df[col] = pd.to_numeric(df[col], errors='coerce')

    df = df.convert_dtypes()

    gdf = add_geometry(df)

    return gdf


def csv(file):
    df = pd.read_csv(file, index_col=0,
                     keep_default_na=False,
                     na_values='')

    # convert the strings to datetimes in UTC. If the strings have timezone information,
    # they will be converted to UTC, otherwise they are assumed to be in UTC.
    df['datetime'] = pd.to_datetime(df['datetime'], utc=True, format='ISO8601')

    gdf = add_geometry(df)
    return gdf


def remote(url):
    data = download(url)
    buf = StringIO(data)

    return csv(buf)


def download(url, chunk_size=1024, json=False):
    buf = StringIO()
    resp = requests.get(url, stream=True)
    total = int(resp.headers.get('content-length', 0))
    with tqdm(desc='downloading', total=total, unit='iB', unit_scale=True, unit_divisor=1024) as bar:
        for data in resp.iter_content(chunk_size=chunk_size):
            size = buf.write(data.decode())
            bar.update(size)
    buf.seek(0)
    if json:
        import json
        return json.loads(buf.read())
    return buf.read()


def add_geometry(df, lat_col='latitude', lon_col='longitude'):
    lats = df[lat_col]
    lons = df[lon_col]
    points = make_points(lons, lats)

    gdf = GeoDataFrame(df, geometry=points, crs="EPSG:4326")
    return gdf


def glm_website_event(ids):
    """Pull individual bolide data from neo-bolide.ndc.nasa.gov by ID

    Downloads light curve and energy data from neo-bolide.ndc.nasa.gov, placing it into
    a dict of lists.

    Parameters
    ----------
    ids: list of str
        List of strings representing the bolide ID's that data are obtained for.
    """

    # dict to store returned data
    data = {'integrated_energy_g16': [], 'integrated_energy_g17': [], 'integrated_energy_g18': [], 'lightcurves': []}

    for bid in tqdm(ids, "Downloading data"):  # for each bolide

        # pull data from website
        web_data = requests.get(API_ENDPOINT_EVENT + bid).json()['data'][0]['attachments']
        row_lcs = []

        # dict to store integrated energies for this event
        # initialize as nan to represent lack of data
        integrated_energies = {'g16': np.nan, 'g17': np.nan, 'g18': np.nan}

        # loop over the attachments returned in the data
        for attachment in web_data:

            # get the satellite ID and geodata
            platform = attachment['platformId']
            geodata = attachment['geoData']

            # since there is data for this satellite, set to zero if nan
            if integrated_energies[platform.lower()] is np.nan:
                integrated_energies[platform.lower()] = 0

            # obtain flux from the geodata, and sum it up to get
            # integrated energy
            flux = [point['energy'] for point in geodata]
            integrated_energies[platform.lower()] += np.sum(flux)

            # obtain times (seconds) from the geodata
            time = [point['time']/1000 for point in geodata]

            # create a LightCurve object
            from astropy.time import Time
            time_obj = Time(time, format='unix')
            lc = LightCurve(time=time_obj, flux=flux)
            lc.meta['MISSION'] = platform
            lc.meta['LABEL'] = platform
            row_lcs.append(lc)

        # create a LightCurveCollection using the invididual LightCurves
        data['lightcurves'].append(LightCurveCollection(row_lcs))

        # enter the integrated energy into the data dict
        for sat, value in integrated_energies.items():
            data[f'integrated_energy_{sat}'].append(value)

    return data



import numpy as np
from geopandas import GeoDataFrame
import pyproj
import cartopy.crs as ccrs
from shapely.geometry import Point, Polygon, LinearRing

from . import GLM_FOV_PATH, ROOT_PATH
from .utils import reconcile_input


def add_boundary(ax, boundary=None, boundary_style={}):
    """Add a boundary to Cartopy GeoAxesSubplot"""

    assert boundary is not None

    # define defaults and add them to the user-defined style,
    # without overriding any user-specified parameters
    boundary_defaults = {"facecolor": "none"}
    boundary_style = reconcile_input(boundary_style, boundary_defaults)

    # get Polygons representing the boundaries
    polygons = get_boundary(boundary)

    # if polygons is not a list, make it a list.
    if type(polygons) is not list:
        polygons = [polygons]

    # define the crs of the polygons
    crs = ccrs.AzimuthalEquidistant(central_latitude=90)
    # add the polygons to the GeoAxesSubplot, passing the boundary style in
    ax.add_geometries(polygons, crs=crs, **boundary_style)


def change_crs(polygons, current, new):
    single_item = False
    if type(polygons) is not list:
        polygons = [polygons]
        single_item = True

    gdf = GeoDataFrame(geometry=polygons, crs=current)
    gdf = gdf.to_crs(new)

    if single_item:
        return gdf.geometry[0]
    else:
        return list(gdf.geometry)


def get_boundary(boundary, collection=True, intersection=False, crs=None):
    """Get specified boundary polygons.

    GOES GLM FOV obtained from a netCDF4 file kindly provided by Katrina Virts
    (NASA/MSFC; katrina.virts@uah.edu)
    based on information from Doug Mach and Clem Tillier.
    Fengyun-4A FOV obtained by eye from several papers.
    Global Meteor Network FOV obtained from KML files at
    https://globalmeteornetwork.org/data/.

    Possible values in boundary:

    - ``'goes'``: Combined FOV of the GLM aboard GOES-16 and GOES-17
    - ``'goes-w'``: GOES-West position GLM FOV, currently corresponding to GOES-17.
      Note that this combines the inverted and non-inverted FOVs.
    - ``'goes-e'``: GOES-East position GLM FOV, currently corresopnding to GOES-16.
    - ``'goes-w-ni'``: GOES-West position GLM FOV, when GOES-17 is not inverted (summer).
    - ``'goes-w-i'``: GOES-West position GLM FOV, when GOES-17 is inverted (winter).
    - ``'goes-17-89.5'``: GOES-17 GLM FOV when it was in its checkout orbit.
    - ``'fy4a'``: Combined FOV of the Fengyun-4A LMI, in both North and South configurations.
    - ``'fy4a-n'``: Fengyun-4A LMI FOV when in the North configuration (summer).
    - ``'fy4a-s'``: Fengyun-4A LMI FOV when in the South configuration (winter).
    - ``'gmn-25km'``: Combined FOV of all Global Meteor Network stations at 25km detection altitude.
    - ``'gmn-70km'``: Combined FOV of all Global Meteor Network stations at 70km detection altitude.
    - ``'gmn-100km'``: Combined FOV of all Global Meteor Network stations at 100km detection altitude.

    Parameters
    ----------
    boundary: str or iterable with multiple strings
        Specifies the boundaries desired. Currently supports:
    collection: bool
        If boundary contains multiple boundaries,
        ``True`` will return a list of boundaries,
        while ``False`` will return their combination (according to the intersection argument)
    intersection: bool
        If boundary contains multiple strings and collection is ``False``, when True will return
        the intersection of the fields-of-view, when False will return their union.

    Returns
    -------
    `~shapely.geometry.Polygon` or list of `~shapely.geometry.Polygon` in the Azimuthal Equidistant
    Coordinate Reference System with a central latitude of 90, central longitude of 0.
    """
    from netCDF4 import Dataset
    from shapely.ops import unary_union

    if type(boundary) is str:
        boundary = boundary.lower()

    valid_boundaries = ['goes-w-ni', 'goes-w-i', 'goes-w', 'goes-e', 'goes',
                        'fy4a-s', 'fy4a-n', 'fy4a', 'goes-17-89.5',
                        'gmn-100km', 'gmn-70km', 'gmn-25km']

    if type(boundary) is str and boundary not in valid_boundaries:
        raise ValueError("unknown boundary \""+boundary+"\". Valid boundaries are: \n"+str(valid_boundaries))

    if type(boundary) is not str and not hasattr(boundary, '__iter__'):
        raise ValueError("boundary must be a string or list")

    if crs is not None:
        polygons = get_boundary(boundary, collection, intersection, crs=None)
        aeqd = pyproj.Proj(proj='aeqd', ellps='WGS84', datum='WGS84', lat_0=90, lon_0=0).srs
        return change_crs(polygons, aeqd, crs)

    if boundary == 'goes-w-ni':
        fov = Dataset(GLM_FOV_PATH, "r", format="NETCDF4")
        lats = fov.variables['G17_fov_lat'][0]
        lons = fov.variables['G17_fov_lon'][0]
        polygon = Polygon(zip(lons, lats))
        return aeqd_from_lonlat(polygon)

    elif boundary == 'goes-w-i':
        fov = Dataset(GLM_FOV_PATH, "r", format="NETCDF4")
        lats = fov.variables['G17_fov_lat_inverted'][0]
        lons = fov.variables['G17_fov_lon_inverted'][0]
        polygon = Polygon(zip(lons, lats))
        return aeqd_from_lonlat(polygon)

    elif boundary == 'goes-17-89.5':
        fov = Dataset(GLM_FOV_PATH, "r", format="NETCDF4")
        lats = fov.variables['G17_fov_lat'][0]
        lons = np.array(fov.variables['G17_fov_lon'][0]) + (-89.5-(-137.2))
        polygon = Polygon(zip(lons, lats))
        return aeqd_from_lonlat(polygon)

    elif boundary == 'goes-w':
        return get_boundary(['goes-w-ni', 'goes-w-i'], collection=False)

    elif boundary == 'goes':
        return get_boundary(['goes-w', 'goes-e'], collection=False)

    elif boundary == 'goes-e':
        fov = Dataset(GLM_FOV_PATH, "r", format="NETCDF4")
        lats = fov.variables['G16_fov_lat'][0]
        lons = fov.variables['G16_fov_lon'][0]
        polygon = Polygon(zip(lons, lats))
        return aeqd_from_lonlat(polygon)

    elif boundary == 'fy4a-n':
        lons = [55, 157.4, 127, 81.75]
        lats = [56.25, 56.25, 14.92, 14.92]
        polygon = fy4a_corners_to_boundary(lons, lats)
        return aeqd_from_lonlat(polygon)

    elif boundary == 'fy4a-s':
        lons = [81.75, 127, 157.4, 55]
        lats = [-14.92, -14.92, -56.25, -56.25]
        polygon = fy4a_corners_to_boundary(lons, lats)
        return aeqd_from_lonlat(polygon)

    elif boundary == 'fy4a':
        return get_boundary(['fy4a-s', 'fy4a-n'], collection=False)

    elif boundary == 'gmn-100km':
        return load_gmn_shp('100km')

    elif boundary == 'gmn-70km':
        return load_gmn_shp('70km')

    elif boundary == 'gmn-25km':
        return load_gmn_shp('25km')

    polygons = []
    for b in boundary:
        new_boundary = get_boundary(b)
        if type(new_boundary) is list:
            polygons += new_boundary
        else:
            polygons.append(new_boundary)

    if collection:
        return polygons
    # either take intersection of FOVs or the union to get a final polygon
    if intersection:
        final_polygon = polygons[0]
        for polygon in polygons:
            final_polygon = final_polygon.intersection(polygon)
    else:
        final_polygon = unary_union(polygons)

    return final_polygon


def aeqd_from_lonlat(polygon):
    # define Azimuthal Equidistant projection
    aeqd = pyproj.Proj(proj='aeqd', ellps='WGS84', datum='WGS84', lat_0=90, lon_0=0).srs
    gdf = GeoDataFrame(geometry=[polygon], crs='epsg:4326')
    gdf = gdf.to_crs(aeqd)
    return gdf.geometry[0]


def load_gmn_shp(distance):
    # load Global Meteor Network shapefiles
    aeqd = pyproj.Proj(proj='aeqd', ellps='WGS84', datum='WGS84', lat_0=90, lon_0=0).srs
    from geopandas import read_file
    path = ROOT_PATH+'/data/gmn/'+distance+'.shp'
    gdf = read_file(path)
    gdf = gdf.to_crs(aeqd)

    # combine them into one polygon, buffering to resolve shape problems
    from shapely.ops import unary_union
    polygons = [p.buffer(0) for p in gdf.geometry]
    return unary_union(polygons)


# get corner points of fy4a FOV and return boundary
def fy4a_corners_to_boundary(lons, lats):

    points = [Point(lon, lat) for lon, lat in zip(lons, lats)]
    gdf = GeoDataFrame(geometry=points, crs='epsg:4326')
    geos = pyproj.Proj(proj='geos', ellps='WGS84', datum='WGS84', h=35785831.0, lon_0=105).srs
    gdf = gdf.to_crs(geos)
    points = gdf.geometry
    linestring = LinearRing(zip([p.x for p in points], [p.y for p in points]))
    length = linestring.length
    points = [linestring.interpolate(x) for x in np.linspace(0, length, 1000)]
    polygon = Polygon(zip([p.x for p in points], [p.y for p in points]))
    gdf = GeoDataFrame(geometry=[polygon], crs=geos)
    gdf = gdf.to_crs('epsg:4326')
    return gdf.geometry[0]


def get_mask(xy, boundary=None, intersection=False):
    # define Azimuthal Equidistant projection
    aeqd = pyproj.Proj(proj='aeqd', ellps='WGS84', datum='WGS84', lat_0=90, lon_0=0).srs

    if boundary is None:
        return np.ones(len(xy))

    polygon = get_boundary(boundary, intersection=intersection, collection=False)

    points = [Point(p[1], p[0]) for p in xy]
    gdf = GeoDataFrame(geometry=points, crs='epsg:4326')
    gdf = gdf.to_crs(aeqd)
    points = gdf.geometry
    points_in_poly = [p.within(polygon) for p in points]
    return np.array(points_in_poly)



API_ENDPOINT_EVENTLIST = "https://neo-bolide.ndc.nasa.gov/service/event/public"
API_ENDPOINT_EVENT = "https://neo-bolide.ndc.nasa.gov/service/event/"
MPLSTYLE = "ggplot"

import pkg_resources
ROOT_PATH = pkg_resources.resource_filename('bolides', '.')
GLM_FOV_PATH = ROOT_PATH + '/data/GLM_FOV_edges.nc'

from .bolide import *
from .bolidelist import *
from .bdf import BolideDataFrame
from .sdf import ShowerDataFrame
from .utils import youtube_photometry
from .constants import GOES_W_LON, GOES_E_LON, GLM_STEREO_MIDPOINT, FY4A_LON

__all__ = ["BolideDataFrame"]
